# SPDX-FileCopyrightText: (C) 2023 Avnet Embedded GmbH
# SPDX-License-Identifier: GPL-3.0-only
"""scotty-test."""

import argparse
import os
import json
import logging
import sys

from colorama import Style
from colorama import Fore

from scotty_test.menu import show_menu
from scotty_test.storage import StorageClient
from scotty_test.runner import Runner
from scotty_test.reporter import Reporter


class ScottyTestLoggingFormatter(logging.Formatter):
    """scotty-test custom logging formatter."""

    format = "%(levelname)s:%(message)s"
    format_exp = "%(levelname)s:%(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: Fore.WHITE + format + Style.RESET_ALL,
        logging.INFO: Fore.WHITE + format + Style.RESET_ALL,
        logging.WARNING: Fore.YELLOW + format + Style.RESET_ALL,
        logging.ERROR: Fore.RED + format + Style.RESET_ALL,
        logging.CRITICAL: Fore.RED + format_exp + Style.RESET_ALL
    }

    def format(self, record):
        """Format a message."""
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def parse_args() -> argparse.Namespace:
    """Argparse handler."""
    parser = argparse.ArgumentParser()
    parser.add_argument('--token', default=os.environ.get('SAS_TOKEN'),
                        help='Azure Storage connection string')
    parser.add_argument('--ghcruser', default=os.environ.get('GHCR_USER'),
                        help='Github username')
    parser.add_argument('--ghcrtoken', default=os.environ.get('GHCR_TOKEN'),
                        help='Github personal access token')
    parser.add_argument('--sdkpath', default='', help='Path to custom SDK')
    parser.add_argument(
        '--sdcard', choices=['manual', 'sdwire'], default='manual', help='SDCard handler')
    return parser.parse_args()


def main():
    """Entry point."""
    # configure logging first
    logger = logging.getLogger('scotty-test')
    handler = logging.StreamHandler()
    handler.setFormatter(ScottyTestLoggingFormatter())
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    args = parse_args()
    if not args.token:
        logging.getLogger('scotty-test').error(
            'You need to pass an Azure blob storage token by either "--token" or via "SAS_TOKEN" environment variable')
        sys.exit(1)
    if not args.ghcruser:
        logging.getLogger('scotty-test').warning(
            '"GHCR_USER" or "--ghcruser" is not set. Some tests will likely fail')
    if not args.ghcrtoken:
        logging.getLogger('scotty-test').warning(
            '"GHCR_TOKEN" or "--ghcrtoken" is not set. Some tests will likely fail')
    _storage = StorageClient(args.token)

    _run_config = None
    _selected_version = None
    while not _run_config:
        _selected_version = show_menu('Select a version', sorted(
            [(x, x) for x in _storage.list_dirs('')]))
        try:
            with open(_storage.download(f'{_selected_version}/build-mapping.json')) as i:
                _run_config = json.load(i)
        except Exception as e:
            logging.getLogger('scotty-test').exception(e)

    # set preferred container version in test
    os.environ['GHCR_CONTAINER_VERSION'] = _selected_version
    # set github container registry credentials
    os.environ['GHCR_USER'] = args.ghcruser or ''
    os.environ['GHCR_TOKEN'] = args.ghcrtoken or ''

    results = []
    while True:
        _selected_machine = show_menu(
            'Select a machine', sorted([x for x in _run_config.keys()]))
        _selected_image = show_menu('Select an image', sorted(
            [(k, {k: v}) for k, v in _run_config[_selected_machine].items()]))

        _runner = Runner(_storage, _selected_version,
                         _selected_machine, _selected_image, args.sdcard, args.sdkpath)
        results.append(_runner.run())

        if not show_menu('Do you want to test another image?', entries=[('Yes', True), ('No', False)]):
            break

    r = Reporter(_storage, _selected_version, results)
    if r.run():
        logging.getLogger('scotty-test').info(
            'Report generated - Please open a PR at https://github.com/avnet-embedded/simplecore-tools')
    else:
        logging.getLogger(
            'scotty-test').debug('Nothing changed since the last report - All good!')


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: (C) 2023 Avnet Embedded GmbH
# SPDX-License-Identifier: GPL-3.0-only

"""Utility collection for scotty-test."""

import subprocess
import os
import tempfile
from typing import Tuple


class Utils():
    """Utils for scotty-test."""

    @staticmethod
    def run_interactive(*args, **kwargs) -> int:
        """Run a command in interactive mode."""
        p = subprocess.Popen(*args, shell=True, **kwargs)
        while True:
            try:
                p.communicate()
            except KeyboardInterrupt:
                pass
            else:
                break

        return p.poll()

    @staticmethod
    def run(*args, **kwargs) -> bool:
        """Run a command and return the exit status."""
        try:
            subprocess.check_call(*args, **kwargs)
            return True
        except subprocess.CalledProcessError:
            return False

    @staticmethod
    def run_output(*args, **kwargs) -> Tuple[bool, str]:
        """Run a command and return the output."""
        try:
            return (True, subprocess.check_output(*args, universal_newlines=True, stderr=subprocess.STDOUT, **kwargs))
        except subprocess.CalledProcessError as e:
            return (False, e.stdout)

    @staticmethod
    def do_in_tempdir(func, *args, **kwargs) -> None:
        """Run code from a temporary directory."""
        ccwd = os.getcwd()
        with tempfile.TemporaryDirectory() as t:
            os.chdir(t)
            func(*args, **kwargs)
        os.chdir(ccwd)



# SPDX-FileCopyrightText: (C) 2023 Avnet Embedded GmbH
# SPDX-License-Identifier: GPL-3.0-only
"""Reporter for scotty-test."""

import json
import logging
import os
import shutil
import subprocess
import tempfile
import time
from typing import Dict

from scotty_test.storage import StorageClient
from scotty_test.utils import Utils


class Reporter():
    """Create a test report."""

    def __init__(self, storage: StorageClient, version: str, results: Dict) -> None:
        """Construct a Reporter instance."""
        self.__storage = storage
        self.__version = version
        self.__results = results
        self.__branch_name = ''

        self.__scotty_path = os.path.join(
            self.__storage.local_storage(), '.scotty')

    def run(self) -> bool:
        """Entry point for Reporter class."""
        self.__setup_scotty()
        self.__create_branch()
        self.__convert_results()
        return self.__git_commit_and_publish()

    def __setup_scotty(self):
        if os.path.exists(self.__scotty_path):
            shutil.rmtree(self.__scotty_path, ignore_errors=True)
        os.makedirs(self.__scotty_path)

        logging.getLogger('scotty-test').info('Setup scotty... please wait')
        Utils.run(['scotty', 'update', '-b', 'kirkstone'],
                  cwd=self.__scotty_path)

    def __convert_results(self):
        for item in self.__results:
            result, testdata = item
            with tempfile.TemporaryDirectory() as t:
                with open(os.path.join(t, 'testresults.json'), 'w') as o:
                    json.dump(result, o)
                with open(os.path.join(t, 'testdata.json'), 'w') as o:
                    json.dump(testdata, o)
                Utils.run(['build/tools/scotty/validation/oeqa-report-enhancer',
                           self.__version,
                           t, 'build/tools/report'],
                          cwd=self.__scotty_path)

    def __create_branch(self):
        self.__branch_name = f'report/{self.__version}-{int(time.time())}'
        Utils.run(['git', 'checkout', '-b', self.__branch_name, 'avnet/kirkstone'],
                  cwd=os.path.join(self.__scotty_path, 'build', 'tools'),
                  stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

    def __git_commit_and_publish(self) -> bool:
        res = Utils.run(['git', 'add', '*'],
                        cwd=os.path.join(self.__scotty_path, 'build', 'tools'),
                        stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        res &= Utils.run(['git', 'commit', '-m', 'report: add test results'],
                         cwd=os.path.join(self.__scotty_path,
                                          'build', 'tools'),
                         stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        res &= Utils.run(['git', 'push', '-u', 'avnet', self.__branch_name],
                         cwd=os.path.join(self.__scotty_path,
                                          'build', 'tools'),
                         stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        return res



# SPDX-FileCopyrightText: (C) 2023 Avnet Embedded GmbH
# SPDX-License-Identifier: GPL-3.0-only
"""Menu provider for scotty-test."""

import inquirer
from inquirer.themes import Default
from blessed import Terminal
from typing import Any, AnyStr, Sequence, Tuple


class AvnetTheme(Default):
    """Avnet specific theming for inquirer."""

    def __init__(self):
        """Init AvnetTheme with our own colors and symbols."""
        super().__init__()

        term = Terminal()

        self.Question.mark_color = term.color_rgb(81, 186, 117)
        self.Question.brackets_color = term.color_rgb(81, 186, 117)
        self.Question.default_color = term.normal
        self.Editor.opening_prompt_color = term.bright_black
        self.Checkbox.selection_color = term.color_rgb(81, 186, 117)
        self.Checkbox.selection_icon = "►"
        self.Checkbox.selected_icon = "[█]"
        self.Checkbox.selected_color = term.color_rgb(81, 186, 117) + term.bold
        self.Checkbox.unselected_color = term.normal
        self.Checkbox.unselected_icon = "[ ]"
        self.List.selection_color = term.color_rgb(81, 186, 117)
        self.List.selection_cursor = "►"
        self.List.unselected_color = term.normal


def show_menu(prompt: AnyStr, entries: Sequence[Tuple[str, Any]], multiple=False) -> Any:
    """Ask the user to make a choice in a list of possibilities."""
    if not entries:
        return ()

    theme = AvnetTheme()

    def as_list(prompt: AnyStr, entries: Sequence[Tuple[str, Any]]):
        questions = [
            inquirer.List('items',
                          message=prompt,
                          choices=entries,
                          default=entries[0],
                          carousel=True
                          )
        ]
        return inquirer.prompt(questions, theme=theme)

    def as_checklist(prompt: AnyStr, entries: Sequence[Tuple[str, Any]]):
        questions = [
            inquirer.Checkbox('items',
                              message=prompt +
                              ' (use SPACE to select, press ENTER to accept)',
                              choices=entries,
                              carousel=True
                              )
        ]
        return inquirer.prompt(questions, theme=theme)

    if multiple:
        answers = as_checklist(prompt, entries)
    else:
        answers = as_list(prompt, entries)

    return answers.get('items', None)


def show_editor(prompt: AnyStr) -> Any:
    """Ask the user to enter text information."""
    theme = AvnetTheme()
    questions = [
        inquirer.Text('items',
                      message=prompt
                      )
    ]
    return inquirer.prompt(questions, theme=theme).get('items', None)



# SPDX-FileCopyrightText: (C) 2023 Avnet Embedded GmbH
# SPDX-License-Identifier: GPL-3.0-only
"""Azure storage functions for scotty-test."""

import os
import logging
from typing import List, Set

from azure.storage.blob import ContainerClient


class StorageClient():
    """Azure storage client."""

    def __init__(self, token: str, *args, **kwargs):
        """Construct the client."""
        self.__token = token
        self.__container = ContainerClient.from_connection_string(
            conn_str=self.__token,
            container_name='blob-sync'
        )

    def local_storage(self) -> str:
        """Return the path to the local storage."""
        return os.path.join(os.environ.get('HOME', '/tmp'), '.local', 'scotty-test')

    def _get_fromlocal_storage(self, path, hash) -> str:
        needle = f'{self.local_storage()}/{path}.{hash}'
        if os.path.exists(needle):
            return needle
        return None

    def _put_tolocal_storage(self, path, hash, stream) -> str:
        needle = f'{self.local_storage()}/{path}.{hash}'
        os.makedirs(os.path.dirname(needle), exist_ok=True)
        with open(needle, 'wb') as o:
            o.write(stream)
        return needle

    def blob_list(self, path) -> List[object]:
        """List blob info."""
        return [x for x in self.__container.list_blobs(name_starts_with=path.lstrip('/') if path else None)]

    def list(self, path: str) -> List[str]:
        """List blob names."""
        return [x.name for x in self.blob_list(path)]

    def list_dirs(self, path: str) -> Set[str]:
        """List directories in the blob storage."""
        return {x.split('/')[0] for x in self.list(path) if '/' in x}

    def download(self, path: str) -> str:
        """Download a blob."""
        expected_hash = self.blob_hash(path)
        res = self._get_fromlocal_storage(path, expected_hash)
        if res is not None:
            return res
        logging.getLogger(
            'scotty-test').info(f'Downloading {path}, please wait...')
        return self._put_tolocal_storage(path,
                                         expected_hash,
                                         self.__container.download_blob(path.lstrip('/')).readall())

    def blob_hash(self, path: str) -> str:
        """Hash of the content of a blob."""
        try:
            return self.blob_list(path)[0].content_settings.content_md5.hex()
        except (IndexError, AttributeError):
            return ''



# SPDX-FileCopyrightText: (C) 2022 Avnet Embedded GmbH
# SPDX-License-Identifier: GPL-3.0-only
# noqa: D100

import setuptools
import os

_version = os.environ.get('SCOTTY_VERSION', '0.0.1')

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

long_description = '''
Scotty is a tool to make the process of building the SimpleCore™ Distro distribution
reproducible on any Linux and Windows computer (compatible with WSL2). It uses a
container to setup the same environment used by our continuous integration
process to ensure that the build on your machine will always be successful for
any of our standard images. Scotty is based on standard open-source tools such as
Docker, repo, ... and mimics the standard bitbake command set (standard tool for
Yocto builds). If you are not familiar with building Yocto BSPs, we strongly
recommend that you use Scotty to start with.

For more details please visit our [Documentation](https://simple.embedded.avnet.com/index.hmtl?link=tools/scotty/README.html).
'''

setuptools.setup(
    author='Avnet Embedded GmbH',
    description='scotty: S(imple)C(ore) O(pen) T(echnology) T(ool for) Y(ou)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='GPL-3.0-only',
    license_files=('LICENSE',),
    entry_points={
        'console_scripts': [
            'bumper = bumper.__main__:main',
            'scotty-test = scotty_test.__main__:main',
        ],
    },
    project_urls={
        'Documentation': 'https://simple.embedded.avnet.com/?link=tools/scotty/README.html',
        'SimpleCore Documentation': 'https://simple.embedded.avnet.com/',
        'Source Code': 'https://github.com/avnet-embedded/simplecore-tools/tree/kirkstone/scotty'
    },
    packages=[
        'bumper',
        'scotty_test',
        'scotty_test.menu',
        'scotty_test.reporter',
        'scotty_test.runner',
        'scotty_test.storage',
        'scotty_test.utils',
    ],
    install_requires=requirements,
    include_package_data=True,
    # As scotty is already in use on PyPi our package is called
    # avnet-scotty
    name='avnet-scotty',
    scripts=[
        'scotty',
        'scotty-runqemu',
        'scripts/vm_bundle.sh',
        'scripts/vm_create.sh.template',
    ],
    version=_version,
)



# SPDX-FileCopyrightText: (C) 2023 Avnet Embedded GmbH
# SPDX-License-Identifier: GPL-3.0-only

"""Test runnner for scotty-test."""

import json
import logging
import os
import re
import time
import zipfile
from typing import Dict, List, Tuple

from scotty_test.menu import show_editor, show_menu
from scotty_test.storage import StorageClient
from scotty_test.utils import Utils


class SDCardHandler():
    """Abstract handler for SDCards."""

    def is_already_flashed(self, image: str) -> bool:
        """Return if the image is already flashed to SD."""
        return False

    def get_menu_options(self) -> List[Tuple[str, str]]:
        """Return the potentialy SD devices."""
        raise NotImplementedError()

    def run(self, device: str, file: str) -> bool:
        """Run the flashing procedure."""
        raise NotImplementedError()


class SDCardManualHandler(SDCardHandler):
    """Handle manual SDCard flashing."""

    def get_menu_options(self) -> List[Tuple[str, str]]:
        """Return the potentialy SD devices."""
        res = set()
        for f in os.listdir('/dev'):
            # match /dev/mmc[0-9] /dev/mmcblk[0-9] and /dev/sd[a-z]
            if re.match(r'^(mmc[0-9]|mmcblk[0-9]|sd[a-z])$', f):
                path = os.path.join('/dev', f)
                res.add((path, path))
        return list(res)

    def run(self, device: str, file: str) -> bool:
        """Run the flashing procedure wih dd."""
        return Utils.run_interactive(
            [f'sudo dd if={file} of={device} bs=64M']) is not None

    def is_already_flashed(self, image: str) -> bool:
        """Return if the image is already flashed to SD."""
        return show_menu(f'Is {image} already flashed to the SDCard?', entries=[('Yes', True), ('No', False)])


class SDCardSDWireHandler(SDCardHandler):
    """Handle SDCard flashing with SDWire."""


class Runner():
    """Test runner implementation."""

    def __init__(self, storage: StorageClient, version: str, machine: str, image_dict: dict, sdhandler: str, sdk: str) -> None:
        """Construct the runner."""
        self.__storage = storage
        self.__machine = machine
        self.__version = version
        self.__sdk_path = sdk
        self.__overlays = set()
        self.__image_name = list(image_dict.keys())[0]
        self.__image_dict = image_dict[self.__image_name]
        self.__sdhandler = SDCardSDWireHandler(
        ) if sdhandler == 'sdwire' else SDCardManualHandler()
        self.__results = {}
        self.__testdata = {}

    def run(self) -> Dict:
        """Entry point for the Runner implementation."""
        Utils.do_in_tempdir(self._prepare_sd)
        Utils.do_in_tempdir(self._do_testrun)
        with open(f'{self.__machine}-{self.__image_name}-{time.time_ns()}-testresults.json', 'w') as o:
            json.dump(self.__results, o)
        return (self.__results, self.__testdata)

    def _unpack_zip(self, path: str):
        file = self.__storage.download(self.__version + path)
        with zipfile.ZipFile(file, 'r') as zip_ref:
            zip_ref.extractall(os.getcwd())
        return os.getcwd()

    def _find_testexport_dir(self, basedir: str, needle: str) -> str:
        for root, dirs, _ in os.walk(basedir):
            for d in dirs:
                if d == needle:
                    return os.path.join(root, d)
        return '/does/not/exist'

    def _find_wic_file(self, basedir: str, needle: str) -> str:
        for root, _, files in os.walk(basedir):
            for f in files:
                if f == needle:
                    return os.path.join(root, f)
        return '/does/not/exist'

    def _find_sdk_installer(self, basedir: str) -> str:
        for root, _, files in os.walk(basedir):
            for f in files:
                if f.endswith('.sh'):
                    return os.path.join(root, f)
        return '/does/not/exist'

    def _find_overlays(self, basedir: str) -> List[str]:
        res = set()
        for _, _, files in os.walk(basedir):
            for f in files:
                if f.endswith('.dtb'):
                    if not f.startswith('overlay-'):
                        continue
                    if f'-{self.__machine}' in f:
                        continue
                    res.add(f)
        return res

    def _configure_overlays(self, selected_overlays: List[str], ip: str, folder: str) -> bool:
        overlay_str = "fdt_overlay=" + " ".join(selected_overlays)

        # patch testdata.json
        with open(os.path.join(folder, 'data', 'testdata.json'), 'r') as f:
            cnt = json.load(f)

        cnt['UBOOT_ENV_VARS'] = overlay_str

        with open(os.path.join(folder, 'data', 'testdata.json'), 'w') as f:
            json.dump(cnt, f, sort_keys=True, indent=2)

        self.__testdata = cnt

        ssh_cmd = ['ssh', '-l', 'root', '-o', 'StrictHostKeyChecking=no',
                   '-o', 'UserKnownHostsFile=/dev/null', ip,
                   f'[ -e /boot/uEnv.txt ] && echo "{overlay_str}" >> /boot/uEnv.txt && /sbin/reboot']
        return Utils.run(ssh_cmd)

    def _prepare_sdk(self) -> Dict:
        def get_sdk_environ(path):
            logging.getLogger(
                'scotty-test').info('Sourcing SDK, please wait...')
            res = {}
            out = Utils.run_output(
                [f'. {path}/environment-setup* > /dev/null && printenv'], shell=True)
            for line in out[1].split('\n'):
                if '=' in line:
                    key, value = line.split('=', 1)
                    res[key] = value
            return res

        if not self.__sdk_path:
            _sdk = self.__image_dict.get('sdk')
            if _sdk:
                _sdk_version = self.__storage.blob_hash(self.__version + _sdk)
                _sdk_local_path = os.path.join(
                    self.__storage.local_storage(), '.sdk', f'{self.__machine}-{_sdk_version}')
                if not os.path.exists(_sdk_local_path):
                    _dir = self._unpack_zip(_sdk)
                    _file = self._find_sdk_installer(_dir)
                    if not os.access(_file, os.X_OK):
                        os.chmod(_file, 0o755)
                    logging.getLogger(
                        'scotty-test').info('Installing SDK, please wait...')
                    Utils.run(f'{_file} -y -d {_sdk_local_path}', shell=True)
                return get_sdk_environ(_sdk_local_path)
        else:
            return get_sdk_environ(self.__sdk_path)
        return {}

    def _handle_hw_map(self, testexport_dir: str) -> None:
        # patch testdata.json
        with open(os.path.join(testexport_dir, 'data', 'testdata.json'), 'r') as f:
            cnt = json.load(f)

        # always patch TOPDIR
        cnt['TOPDIR'] = testexport_dir

        # set interactive
        cnt['INTERACTIVE_TEST'] = '1'

        # find HW_TESTING_MAP
        hw_map = json.loads(cnt.get('HW_TESTING_MAP_ORIG', '{}'))
        hw_map = {k: v for k, v in hw_map.items()}
        if any(k for k, v in hw_map.items() if v != -1):
            hw_map_selection = show_menu("Please choose the features your base board offers", [
                                         (k, k) for k, v in hw_map.items() if v != -1], True)
        for item in hw_map_selection:
            hw_map[item] = 1

        cnt['HW_TESTING_MAP'] = json.dumps(hw_map, sort_keys=True)

        with open(os.path.join(testexport_dir, 'data', 'testdata.json'), 'w') as f:
            json.dump(cnt, f, sort_keys=True, indent=2)

        self.__testdata = cnt

    def _do_testrun(self) -> Dict:
        _cur_environ = os.environ.copy()
        logging.getLogger('scotty-test').info('Preparing the SDK...')
        os.environ.update(self._prepare_sdk())
        _dir = self._unpack_zip(self.__image_dict.get('testexport'))
        _folder = self._find_testexport_dir(_dir, self.__image_name)
        self._handle_hw_map(_folder)
        ip = show_editor('Power up the board and tell us the IP')
        if self.__overlays:
            _selected_overlays = show_menu(
                "Please select the matching overlays", [(x, x) for x in self.__overlays], True)
        elif self.__machine in ['sm2s-intel-all']:
            _selected_overlays = [show_menu(
                "What base board are you using?", [
                    ('EP5-001', 'overlay-baseboard-ep5-gpioexpander.dtb'),
                    ('EP5-002', 'overlay-baseboard-ep5.dtb'),
                    ('EP1', 'overlay-baseboard-ep1.dtb'),
                ])]
        else:
            _selected_overlays = [show_editor("Enter the to be used overlays")]
        self._configure_overlays(_selected_overlays, ip, _folder)

        logging.getLogger('scotty-test').info('Waiting for reboot...')
        time.sleep(15.0)

        os.chmod(os.path.join(_folder, 'oe-test'), 0o755)
        os.chdir(_folder)
        logging.getLogger('scotty-test').info('Starting tests...')
        Utils.run_interactive(f'./oe-test runtime --target-ip={ip}')
        with open(os.path.join(_folder, 'runtime-results/testresults.json')) as i:
            self.__results = json.load(i)
        os.environ = _cur_environ

    def _prepare_sd(self):
        if not self.__sdhandler.is_already_flashed(self.__image_name):
            _selected_device = None
            while not _selected_device:
                _selected_device = show_menu(
                    'Please select the SDCard device', entries=self.__sdhandler.get_menu_options() + [('Rescan', None)])
            _dir = self._unpack_zip(self.__image_dict.get('image'))
            _file = self._find_wic_file(
                _dir, f'{self.__image_name}-{self.__machine}.wic')
            logging.getLogger('scotty-test').info('Flashing SDCard...')
            while not self.__sdhandler.run(_selected_device, _file):
                pass
            self.__overlays = self._find_overlays(_dir)
